import { MatcherException } from '../../exceptions/matcher.exception';
import { MarginDurations } from '../types/margin-durations.type';
import { IRequestTime } from '../interfaces/time-request.interface';
import { TimingDays, TimingFrequency, Days } from '../types/timing';
import { Schedule } from '../types/schedule.type';

export class Time {
  _timeRequest: IRequestTime;
  _defaultMarginDuration: number;
  _defaultValidityDuration: number;
  frequency: TimingFrequency;
  fromDate: Date;
  toDate: Date;
  schedule: Schedule;
  marginDurations: MarginDurations;

  constructor(
    timeRequest: IRequestTime,
    defaultMarginDuration: number,
    defaultValidityDuration: number,
  ) {
    this._timeRequest = timeRequest;
    this._defaultMarginDuration = defaultMarginDuration;
    this._defaultValidityDuration = defaultValidityDuration;
    this.schedule = {};
    this.marginDurations = {
      mon: defaultMarginDuration,
      tue: defaultMarginDuration,
      wed: defaultMarginDuration,
      thu: defaultMarginDuration,
      fri: defaultMarginDuration,
      sat: defaultMarginDuration,
      sun: defaultMarginDuration,
    };
  }

  init() {
    this._validateBaseDate();
    this._validatePunctualRequest();
    this._validateRecurrentRequest();
    this._setPunctualRequest();
    this._setRecurrentRequest();
    this._setMargindurations();
  }

  _validateBaseDate() {
    if (!this._timeRequest.departure && !this._timeRequest.fromDate) {
      throw new MatcherException(3, 'departure or fromDate is required');
    }
  }

  _validatePunctualRequest() {
    if (this._timeRequest.departure) {
      this.fromDate = this.toDate = new Date(this._timeRequest.departure);
      if (!this._isDate(this.fromDate)) {
        throw new MatcherException(3, 'Wrong departure date');
      }
    }
  }

  _validateRecurrentRequest() {
    if (this._timeRequest.fromDate) {
      this.fromDate = new Date(this._timeRequest.fromDate);
      if (!this._isDate(this.fromDate)) {
        throw new MatcherException(3, 'Wrong fromDate');
      }
    }
    if (this._timeRequest.toDate) {
      this.toDate = new Date(this._timeRequest.toDate);
      if (!this._isDate(this.toDate)) {
        throw new MatcherException(3, 'Wrong toDate');
      }
      if (this.toDate < this.fromDate) {
        throw new MatcherException(3, 'toDate must be after fromDate');
      }
    }
    if (this._timeRequest.fromDate) {
      this._validateSchedule();
    }
  }

  _validateSchedule() {
    if (!this._timeRequest.schedule) {
      throw new MatcherException(3, 'Schedule is required');
    }
    if (
      !Object.keys(this._timeRequest.schedule).some((elem) =>
        Days.includes(elem),
      )
    ) {
      throw new MatcherException(3, 'No valid day in the given schedule');
    }
    Object.keys(this._timeRequest.schedule).map((day) => {
      const time = new Date('1970-01-01 ' + this._timeRequest.schedule[day]);
      if (!this._isDate(time)) {
        throw new MatcherException(3, `Wrong time for ${day} in schedule`);
      }
    });
  }

  _setPunctualRequest() {
    if (this._timeRequest.departure) {
      this.frequency = TimingFrequency.FREQUENCY_PUNCTUAL;
      this.schedule[TimingDays[this.fromDate.getDay()]] =
        this.fromDate.getHours() + ':' + this.fromDate.getMinutes();
    }
  }

  _setRecurrentRequest() {
    if (this._timeRequest.fromDate) {
      this.frequency = TimingFrequency.FREQUENCY_RECURRENT;
      if (!this.toDate) {
        this.toDate = this._addDays(
          this.fromDate,
          this._defaultValidityDuration,
        );
      }
      this._setSchedule();
    }
  }

  _setSchedule() {
    Object.keys(this._timeRequest.schedule).map((day) => {
      this.schedule[day] = this._timeRequest.schedule[day];
    });
  }

  _setMargindurations() {
    if (this._timeRequest.marginDuration) {
      const duration = Math.abs(this._timeRequest.marginDuration);
      this.marginDurations = {
        mon: duration,
        tue: duration,
        wed: duration,
        thu: duration,
        fri: duration,
        sat: duration,
        sun: duration,
      };
    }
    if (this._timeRequest.marginDurations) {
      if (
        !Object.keys(this._timeRequest.marginDurations).some((elem) =>
          Days.includes(elem),
        )
      ) {
        throw new MatcherException(
          3,
          'No valid day in the given margin durations',
        );
      }
      Object.keys(this._timeRequest.marginDurations).map((day) => {
        this.marginDurations[day] = Math.abs(
          this._timeRequest.marginDurations[day],
        );
      });
    }
  }

  _isDate = (date: Date): boolean => {
    return date instanceof Date && isFinite(+date);
  };

  _addDays = (date: Date, days: number): Date => {
    const result = new Date(date);
    result.setDate(result.getDate() + days);
    return result;
  };
}
